#include "postgres.h"

#include "miscadmin.h"
#include "postmaster/bgworker.h"
#include "postmaster/interrupt.h"
#include "storage/latch.h"
#include "executor/spi.h"
#include "tcop/tcopprot.h"
#include "utils/wait_classes.h"

PG_MODULE_MAGIC;

extern PGDLLEXPORT void TestExMain(Datum main_arg);

static void
do_work()
{
	int ret;
	MemoryContext oldcontext;

	StartTransactionCommand();
	PushActiveSnapshot(GetTransactionSnapshot());

	oldcontext = CurrentMemoryContext;

	PG_TRY();
	{

		if ((ret = SPI_connect()) != SPI_OK_CONNECT)
       			elog(ERROR, "SPI_connect failed: %d", ret);

		if ((ret = SPI_execute("SELECT * FROM TEST", false, 0)) < 0)
       			elog(ERROR, "SPI_exec failed: %d", ret);

		SPI_finish();

		PopActiveSnapshot();

                CommitTransactionCommand();
	}
	PG_CATCH();
        {
                ErrorData  *edata;

                MemoryContextSwitchTo(oldcontext);

                edata = CopyErrorData();
                elog(WARNING, "%s", edata->message);
                FreeErrorData(edata);
                FlushErrorState();

                AbortCurrentTransaction();
        }
        PG_END_TRY();

	pg_usleep(10000);
//	CommitTransactionCommand();
}



void
TestExMain(Datum main_arg)
{
	pqsignal(SIGHUP, SignalHandlerForConfigReload);
	pqsignal(SIGTERM, die);
	BackgroundWorkerUnblockSignals();

	BackgroundWorkerInitializeConnection("postgres", NULL, 0);

	for (;;)
	{
		int			rc;

		CHECK_FOR_INTERRUPTS();

		if (ConfigReloadPending)
		{
			ConfigReloadPending = false;
			ProcessConfigFile(PGC_SIGHUP);
		}

		do_work();

		ResetLatch(MyLatch);
		rc = WaitLatch(MyLatch,
					   WL_LATCH_SET | WL_TIMEOUT | WL_POSTMASTER_DEATH,
					   1000,
					   PG_WAIT_EXTENSION);

		if (rc & WL_POSTMASTER_DEATH || ProcDiePending)
			break;
	}
}

static void
TestExRegisterWorker(void)
{
	BackgroundWorker bgw;

	memset(&bgw, 0, sizeof(bgw));

	bgw.bgw_flags = BGWORKER_SHMEM_ACCESS |
		BGWORKER_BACKEND_DATABASE_CONNECTION;
	bgw.bgw_start_time = BgWorkerStart_RecoveryFinished;

	snprintf(bgw.bgw_library_name, BGW_MAXLEN, "testex");
	snprintf(bgw.bgw_function_name, BGW_MAXLEN, "TestExMain");
	bgw.bgw_restart_time = 1;
	bgw.bgw_notify_pid = 0;

	bgw.bgw_main_arg = (Datum) NULL;
	RegisterBackgroundWorker(&bgw);
}

void
_PG_init()
{
	if (!process_shared_preload_libraries_in_progress)
		elog(ERROR, "module can be loaded only on startup");

	TestExRegisterWorker();
}
